/*
 * Licensed Materials - Property of IBM                               
 * Component IBM Rational ROSE_RT BUILD : TARGET_RTS                  
 * Copyright IBM Corporation 1999, 2008. All Rights Reserved.       
 * U.S. Government Users Restricted Rights -  Use, duplication or     
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

#include <RTActor_class.h>
#include <RTDebugger.h>
#include <RTEventInfo.h>
#include <RTMessage.h>
#include <RTObserver.h>
#include <RTPortDescriptor.h>
#include <RTProtocolDescriptor.h>

static void adjustEvent( RTEventInfo * eventInfo )
{
    // This method will obtain the signal name if the event kind is one of
    // those that will use it later. The reason it has to be done now is that
    // the sending or receiving end (including its port) might not exist
    // later when the event is reported back to the observer (Toolset).

    const RTMessage * msg = eventInfo->message;

    eventInfo->signalName = "?";

    switch( eventInfo->eventKind )
    {
	// simple cases - just call getSignalName()
	case RTEventInfo::EventChain:
	case RTEventInfo::EventState:
	case RTEventInfo::EventMsgIn:
	case RTEventInfo::EventMsgOutIn:
	case RTEventInfo::EventTimer:
	    eventInfo->signalName = msg->getSignalName();
	    break;

	// slightly more complicated - get signal from sender
	case RTEventInfo::EventMsgOut:
	{
	    const RTActor_class    * info = eventInfo->info;
	    const RTPortDescriptor * port = info->ports;

	    for( int numPorts = info->numPorts; --numPorts >= 0; ++port )
	    {
		if( port->id != msg->_sendPortId )
		    continue;

		const RTProtocolDescriptor * protocol = port->protocol;

		if( protocol == (const RTProtocolDescriptor *)0 )
		    return;

		eventInfo->signalName =
		    protocol->outSignalName( msg->getSignal() );
		return;
	    }

	    break;
	}

#ifdef _lint
	// these events don't use the signal name
	case RTEventInfo::EventPresence:
	case RTEventInfo::EventReference:
	case RTEventInfo::EventFSM:
	case RTEventInfo::EventLateFSM:
	case RTEventInfo::EventVariable:
	case RTEventInfo::EventHistory:
	case RTEventInfo::EventControl:
#endif
	default:
	    break;
    }
}

int RTDebugger::addEvent( RTEventInfo * eventInfo )
{
    int  shouldHalt = 0;
    bool shouldSave = false;

    for( int iObs = 0; iObs < observers.size(); ++iObs )
    {
	if( eventInfo->observerId >= 0 &&
	    eventInfo->observerId != iObs )
	{
	    continue;
	}

	RTObserver * observer = (RTObserver *)observers[ iObs ];

	if( observer == (RTObserver *)0 )
	    continue;

	// -1: not interested, 0: interested, 1: halt
	int result = observer->prepareEvent( *eventInfo );

	if( result >= 0 )
	{
	    shouldSave = true;

	    if( result > 0 )
	    {
		shouldHalt = 1;
		break;
	    }
	}
    }

    if( ! shouldSave )
    {
	delete eventInfo;
	return 0;
    }

    adjustEvent( eventInfo );

    if( eventHead == (RTEventInfo *)0 ) // event list is empty; wakeup needed
    {
	eventHead = eventTail = eventInfo;
#if USE_THREADS
	wakeupUdp();
#endif
    }
    else // at least one event in list; wakeup not needed
    {
	eventInfo->next = eventHead;
	eventHead->prev = eventInfo;
	eventHead       = eventInfo;
    }

    return shouldHalt;
}
