/*
 * Licensed Materials - Property of IBM                               
 * Component IBM Rational ROSE_RT BUILD : TARGET_RTS                  
 * Copyright IBM Corporation 1999, 2008. All Rights Reserved.       
 * U.S. Government Users Restricted Rights -  Use, duplication or     
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

#include <RTDiag.h>
#include <RTMain.h>
#include <RTThreadInfo.h>
#include <RTpSOS.h>

#include <psos.h>
#include <prepc.h>
#include <pna.h>
#include <version.h>
#include <cxx/init_cxx.h>

static void InitializeDevices( void )
{
    unsigned long   rc;
    unsigned long   ioretval;
    void          * data;

    if( (rc = de_init( DEV_SERIAL, (void *)0, &ioretval, &data )) != 0 )
	k_fatal( 0x10000 + rc, 0 );

    if( (rc = de_init( DEV_TIMER, (void *)0, &ioretval, &data )) != 0 )
	k_fatal( 0x10000 + rc, 0 );
}

extern "C" void root( void )
{
    // initialize C++ runtime stuff
    InitCxxRuntime();

    InitializeDevices();

    (void)RTMain::entryPoint( 0, (const char * const *)0 );

    rt_thread_exit();
}

// We cannot use t_delete to destroy threads arbitrarily because pSOS
// will reclaim all the memory they allocated even though other threads
// may still be using some of that memory: we must just suspend them.
// The root task restarts them with either another job or a request
// to terminate.

RTTaskManager * RTTaskManager::soleInstance;

RTTaskManager::RTTaskManager( void ( * entry )( void * ),
			      void   * arg,
			      int      priority,
			      int      stack_size )
    : mutex  (                   )
    , sync   (                   )
    , tasks  ( (RTTask       *)0 )
    , running(                 0 )
    , creates( (RTTaskCreate *)0 )
    , joins  ( (RTTaskJoin   *)0 )
{
    soleInstance = this;

    mutex.enter();
    (void)assignTask( entry, arg, priority, stack_size );
    mutex.leave();
}

RTTaskManager::~RTTaskManager( void )
{
    while( running != 0 )
    {
	sync.wait();

	mutex.enter();

	// Do joins first so we are more likely to have a thread we can reuse.
	while( joins != (RTTaskJoin *)0 )
	{
	    RTTaskJoin * request = joins;

	    joins = request->next;

	    request->next = (RTTaskJoin *)0;

	    RTTask * task = request->task;

	    switch( task->state )
	    {
	    case RTTask::Running:
		RTAssert( task->waiter == (RTTaskJoin *)0 );
		task->waiter = request;
		break;

	    case RTTask::Zombie:
		task->state   = RTTask::Reusable;
		request->task = (RTTask *)0;
		request->sync.signal();
		break;

	    case RTTask::Reusable:
	    default:
		RTDiag::panic( "RTTaskManager:: bad task state" );
		break;
	    }
	}

	while( creates != (RTTaskCreate *)0 )
	{
	    RTTaskCreate * request = creates;

	    creates = request->next;

	    request->next = (RTTaskCreate *)0;

	    request->task = assignTask( request->entry,
					request->arg,
					request->priority,
					request->stack_size );
	    request->sync.signal();
	}

	mutex.leave();
    }

    unsigned long args[ 4 ];

    args[ 0 ] = 0UL;
    args[ 1 ] = 0UL;
    args[ 2 ] = 0UL;
    args[ 3 ] = 0UL;

    for( RTTask * task; (task = tasks) != (RTTask *)0; )
    {
	if( (tasks = task->next) == task )
	    tasks = (RTTask *)0;
	else
	{
	    task->next->prev = task->prev;
	    task->prev->next = task->next;
	}

	// instruct the task to delete itself
	if( t_restart( task->tid, args ) != 0UL )
	    RTDiag::panic( "t_restart" );

	delete task;
    }
}

RTTask * RTTaskManager::startTask( void ( * entry )( void * ),
				   void   * arg,
				   int      priority,
				   int      stack_size )
{
    RTTaskCreate request;

    request.task       = (RTTask *)0;
    request.entry      = entry;
    request.arg        = arg;
    request.priority   = (unsigned long)priority;
    request.stack_size = (unsigned long)stack_size;

    RTTaskManager * manager = soleInstance;

    manager->mutex.enter();

    request.next     = manager->creates;
    manager->creates = &request;

    manager->mutex.leave();

    manager->sync.signal();

    while( request.task == (RTTask *)0 )
	request.sync.wait();

    return request.task;
}

void RTTaskManager::joinTask( RTTask * task )
{
    RTTaskJoin request;

    request.task = task;

    RTTaskManager * manager = soleInstance;

    manager->mutex.enter();

    request.next   = manager->joins;
    manager->joins = &request;

    manager->mutex.leave();

    manager->sync.signal();

    while( request.task != (RTTask *)0 )
	request.sync.wait();
}

RTTask * RTTaskManager::assignTask( void          ( * entry )( void * ),
				    void            * arg,
				    unsigned long     priority,
				    unsigned long     stack_size )
{
    RTTask        * task;
    unsigned long   args[ 4 ];
    unsigned long   old_priority;

    args[ 0 ] = 0UL;
    args[ 1 ] = (unsigned long)entry;
    args[ 2 ] = (unsigned long)arg;
    args[ 3 ] = 0UL;

    if( (task = tasks) != (RTTask *)0 )
    {
	RTTask * best = (RTTask *)0;

	do
	{
	    if( task->state      == RTTask::Reusable &&
		task->stack_size >= stack_size )
	    {
		if( best == (RTTask *)0 )
		    best = task;
		else if( task->stack_size < best->stack_size )
		    best = task;
	    }
	}
	while( (task = task->next) != tasks );

	task = best;
    }

    if( task != (RTTask *)0 )
    {
	task->state = RTTask::Running;
	running += 1;

	args[ 0 ] = (unsigned long)task;

	if( t_setpri( task->tid, priority, &old_priority ) != 0UL )
	    RTDiag::panic( "t_setpri" );

	else if( t_restart( task->tid, args ) != 0UL )
	    RTDiag::panic( "t_restart" );
    }
    else
    {
	static char   task_name[] = "rose";
	unsigned long tid         = 0UL;

	if( t_create( task_name, priority, 8192UL, stack_size,
		      T_LOCAL | T_FPU, &tid ) != 0UL )
	    RTDiag::panic( "t_create" );

	else
	{
	    task = new RTTask;

	    if( tasks == (RTTask *)0 )
	    {
		task->next = task;
		task->prev = task;
		tasks      = task;
	    }
	    else
	    {
		task->next        = tasks;
		task->prev        = tasks->prev;
		tasks->prev->next = task;
		tasks->prev       = task;
	    }

	    task->tid        = tid;
	    task->stack_size = stack_size;
	    task->state      = RTTask::Running;
	    task->waiter     = (RTTaskJoin *)0;
	    running += 1;

	    args[ 0 ] = (unsigned long)task;

	    if( t_start( tid, T_PREEMPT | T_TSLICE | T_NOASR | T_USER,
			      (void (*)( void ))wrapper, args ) != 0UL )
	    {
		delete task;
		task = (RTTask *)0;
		(void)t_delete( tid );
		RTDiag::panic( "t_start" );
	    }
	}
    }

    return task;
}

void RTTaskManager::wrapper( RTTask   * task,
			     void   ( * entry )( void * ),
			     void     * arg )
{
    if( task == (RTTask *)0 )
	rt_thread_exit();
    else
    {
	(*entry)( arg );

	RTTaskManager * manager = soleInstance;

	manager->mutex.enter();

	RTTaskJoin * waiter = task->waiter;

	if( waiter == (RTTaskJoin *)0 )
	    task->state = RTTask::Zombie;
	else
	{
	    task->state  = RTTask::Reusable;
	    task->waiter = (RTTaskJoin *)0;

	    waiter->task = (RTTask *)0;
	    waiter->sync.signal();
	}
	manager->running -= 1;

	manager->mutex.leave();
	manager->sync.signal();

	for(;;)
	    if( t_suspend( 0 ) != 0UL )
		RTDiag::panic( "t_suspend" );
    }
}

void rt_thread_exit( void )
{
    fclose( (FILE *)0 ); // release pREPC+ resources

    close( 0 ); // release pNA+ resources

#if defined( VERSION ) && VERSION - 0 >= 220
    free( (void *)-1 ); // release pNA+ memory
#endif

    if( t_delete( 0 ) != 0UL )
	RTDiag::panic( "t_delete" );
}
